/*
   multiply a N x N matrix with a vector;

   Input format:

   Input ::= Dimension Matrix Vector
   Dimension ::= Digits
   Matrix ::= { FloatingPointNumber } # Dimension x Dimension times
   Vector ::= { FloatingPointNumber } # Dimension times

   Output format:

   Output ::= { FloatingPointNumber } # Dimension times
*/
#include <mpi.h>
#include <iostream>
#include <cassert>

using namespace std;

enum { NEXT_ROW, FINISH };

static void
gemv_master(int n, double** A, double *x, double* y,
            int nofslaves) {
   // broadcast parameters that are required by all slaves
   MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
   MPI_Bcast(x, n, MPI_DOUBLE, 0, MPI_COMM_WORLD);

   // send out initial tasks for all slaves
   int* tasks = new int[nofslaves]; // remember the task for each of the slaves
   int next_task = 0;
   for (int slave = 1; slave <= nofslaves; ++slave) {
      if (next_task < n) {
         int row = next_task++; // pick next remaining task
         MPI_Send(A[row], n, MPI_DOUBLE, slave, NEXT_ROW, MPI_COMM_WORLD);
         tasks[slave-1] = row; // remember which task was sent out to whom
      } else {
         // there is no work left for this slave
         MPI_Send(0, 0, MPI_DOUBLE, slave, FINISH, MPI_COMM_WORLD);
      }
   }

   // collect results and send out remaining tasks
   int done = 0;
   while (done < n) {
      // receive result of a completed task
      double value = 0; // initialize it to get rid of warning
      MPI_Status status;
      MPI_Recv(&value, 1, MPI_DOUBLE,
         MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      int slave = status.MPI_SOURCE;
      int row = tasks[slave-1];
      y[row] = value;
      ++done;
      // send out next task, if there is one left
      if (next_task < n) {
         row = next_task++;
         MPI_Send(A[row], n, MPI_DOUBLE, slave, NEXT_ROW, MPI_COMM_WORLD);
         tasks[slave-1] = row;
      } else {
         // send notification that there is no more work to be done
         MPI_Send(0, 0, MPI_DOUBLE, slave, FINISH, MPI_COMM_WORLD);
      }
   }

   // release allocated memory
   delete[] tasks;
}

static void gemv_slave() {
   int n;
   MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
   double* x = new double[n];
   MPI_Bcast(x, n, MPI_DOUBLE, 0, MPI_COMM_WORLD);

   double* row = new double[n];
   // receive tasks and process them
   for(;;) {
      // receive next task
      MPI_Status status;
      MPI_Recv(row, n, MPI_DOUBLE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      if (status.MPI_TAG == FINISH) break;
      // process it
      double result = 0;
      for (int i = 0; i < n; ++i) {
         result += row[i] * x[i];
      }
      // send result back to master
      MPI_Send(&result, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
   }
   // release allocated memory
   delete[] x; delete[] row;
}

static bool read_parameters(int& n, double**& matrix, double*& vector) {
   if (!(cin >> n) || n <= 0) return false;
   matrix = new double*[n];
   for (int i = 0; i < n; ++i) {
      matrix[i] = new double[n];
      for (int j = 0; j < n; ++j) {
         if (!(cin >> matrix[i][j])) return false;
      }
   }
   vector = new double[n];
   for (int i = 0; i < n; ++i) {
      if (!(cin >> vector[i])) return false;
   }
   return true;
}

int main(int argc, char** argv) {
   MPI_Init(&argc, &argv);

   int rank; MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   int nofslaves; MPI_Comm_size(MPI_COMM_WORLD, &nofslaves);
   --nofslaves; assert(nofslaves > 0);

   if (rank == 0) {
      int n; double** A; double* x;
      if (!read_parameters(n, A, x)) {
         cerr << "Invalid input!" << endl;
         MPI_Abort(MPI_COMM_WORLD, 1);
      }
      double* y = new double[n];
      gemv_master(n, A, x, y, nofslaves);
      for (int i = 0; i < n; ++i) {
         cout << " " << y[i] << endl;
      }
   } else {
      gemv_slave();
   }

   MPI_Finalize();
}
